/*
 * linux/arch/arm/drivers/sound/voice.c
 *
 * Copyright (C) 1994, 1995, 1996 Russell King
 *
 * Voice driver
 */

#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include "soundcont.h"
#include "soundinit.h"
#include "soundirq.h"
#include "voice.h"

extern const struct voice_driver voice_sample;
extern const struct voice_driver voice_wave;

/*
 * Allocate a voice driver of type 'type', passing 'data' to it
 */

int voice_allocate(int type,int data)
{
	switch(type)
	{
		case VOICE_SAMPLE: return voice_sample.vd_alloc(data);
		case VOICE_WAVE:   return voice_wave.vd_alloc(data);
		default:	   return -ENODEV;
	}
}

/*
 * Reallocate the buffer associated with driver 'vd'
 *
 * This routine can be called while the driver is in use,
 * however, it may wait...
 *
 * If unsuccessful, returns -ENOMEM, but sample will keep playing
 *
 */

int voice_realloc(struct voice_driver *vd, int newlength)
{
	char *nb,*ob;
	int oldlength;
	unsigned long flags;

	nb = (char *)kmalloc(newlength, GFP_KERNEL);
	if(!nb)
		return -ENOMEM;

/* Now for the trick */
	save_flags(flags);
	cli();
	ob = vd->vd_buf;
	oldlength=vd->vd_size;
	if(vd->vd_tail<vd->vd_head)
	{/* Simple case */
		memcpy(nb, ob+vd->vd_tail, vd->vd_cnt);
		vd->vd_buf=nb;
		vd->vd_size=newlength;
	}
	else
	{/* Complicated case - copy top part to top of new buffer, then same down the bottom */
		int newtail=newlength+vd->vd_tail-oldlength;
		memcpy(nb+newtail, ob+vd->vd_tail, oldlength-vd->vd_tail);
		memcpy(nb, ob, vd->vd_head);
		vd->vd_buf=nb;
		vd->vd_tail=newtail;
		vd->vd_size=newlength;
	}
	kfree_s(ob,oldlength);
	restore_flags(flags);
	return 0;
}

/*
 * Install a voice on a channel
 */

int voice_install(struct voice_driver *vd, int channel)
{
	struct voice *vdrv;

	if(vd->vd_channel != -1)
		return -EINVAL;
	vd->vd_channel = channel;
	vdrv = (struct voice *)kmalloc(sizeof(struct voice), GFP_KERNEL);
	if(!vdrv)
		return -ENOMEM;
	
	vd->vd_install(vd, vdrv);

	sound_voice[channel] = vdrv;
	sound_allocated[channel] = 1;
	soundirq_inc();
	return 0;
}

/* 
 * find a free channel or -1 on error (channel = -1 to find, channel no to check)
 */

int voice_findfree(int channel)
{
	for(; channel < 8; channel++)
		if(!sound_allocated[channel])
			return channel;
	return -1;
}

/*
 * Remove a voice from a channel
 */

void voice_remove(struct voice_driver *vd)
{
	struct voice *vdrv;

	if(vd->vd_channel == -1)
		return;
	vdrv = sound_voice[vd->vd_channel];
	
	if(vd->vd_remove)
		vd->vd_remove(vd, vdrv);
	
	sound_voice[vd->vd_channel] = &sound_to_quiet_voice;
	sound_allocated[vd->vd_channel] = 0;
	vd->vd_channel = -1;

	kfree_s(vdrv, sizeof(struct voice));
}

/*
 * Release a voice, and return its memory
 */

int voice_release(struct voice_driver *vd)
{
	struct voice *vdrv;

	if(vd->vd_channel != -1)
	{
		vdrv = sound_voice[vd->vd_channel];

		if(vd->vd_remove)
			vd->vd_remove(vd, vdrv);

		sound_voice[vd->vd_channel] = &sound_to_quiet_voice;
		sound_allocated[vd->vd_channel] = 0;
		vd->vd_channel = -1;

		kfree_s(vdrv, sizeof(struct voice));
	}

	if(vd->vd_buf)
		kfree_s(vd->vd_buf,vd->vd_size);
	kfree_s(vd, sizeof(struct voice_driver));
	return 0;
}

/*
 * Change the volume on a channel
 */

int voice_volume(struct voice_driver *vd,int newvolume)
{
	if(vd->vd_vol)
		return vd->vd_vol(vd,newvolume);
	if(newvolume == -1)
		return vd->vd_volume;
	vd->vd_volume=newvolume;
	return 0;
}
